{
Object based handling of application and icons for PortableApps menu
handles multiple level categories and custom sorting.

Copyright 2007 Karl McMurdo

Portions Copyright 2006-2007 John T. Haller of PortableApps.com

Please see the included license file for the full details on
all included components.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


Todo: Adjust variable naming to match main portableappsmenu program style
      Add comments to all procedures/functions
      Different Icons for open and closed menus
      Rename Categories
      Custom dialog for exe (alternate hint and exit program/params)
      Custom Dialog for ctg (custom icons and Hint)
}
unit AppObjects;

interface

uses
  Windows, Classes, Graphics, Controls, Forms, Buttons, Menus, SysUtils, IniFiles, ShellAPI;

type
  TAppItem = class(TObject)
  private
    fName:           String;
    fCaption:        String;
    fExePath:        String;
    fHint:           String;
    fIconHandle:     HIcon;
    fOpenIconHandle: HIcon;
    fExecOnClose:    Boolean;
    fAutoExec:       Boolean;
    fVisible:        Boolean;
    fSortOrder:      Integer;
    fLevel:          Integer;
    fCategoryName:   String;
    fIsMenu:         Boolean;
    fMenuOpen:       Boolean;
    fExitExe:        String;  // App to run on menu exit IF app has run this session.
    fExitParams:     String;
    fHasRun:         Boolean; // Indicates app has run since menu started
    fIsUpdating:     Boolean; // Indicates more than one option is being updated so don't update paf.
    fAppList:        TObject; // So app can easily reference parent
    procedure SetCaption(strValue: String);
    procedure SetHint(strValue: String);
    procedure SetExecOnClose(bolValue: Boolean);
    procedure SetAutoExec(bolValue: Boolean);
    procedure SetVisible(bolValue: Boolean);
    procedure SetCategoryName(strValue: String);
    procedure SetExitExe(strValue: String);
    procedure SetExitParams(strValue: String);
    procedure SetIsUpdating(bolValue: Boolean);
    function GetSortString: String;
    constructor Create(intDBID: Integer; DB: TIniFile; IconStream: TStream); overload;
    constructor Create(ExePath: String); overload;
    procedure SaveToDB(DBID: Integer; DB: TIniFile; IconStream: tStream);
    procedure SetModified; // Sets Modified flag in parent applist if applist is assigned
    procedure AddOpenIcon(strIcoPath: String);
    procedure UpdateSort;
  public
    property Name: String Read fName Write fName;
    property Caption: String Read fCaption Write SetCaption;
    property ExePath: String Read fExePath Write fExePath;
    property Hint: String Read fHint Write SetHint;
    property IconHandle: HIcon Read fIconHandle Write fIconHandle;
    property OpenIconHandle: HIcon Read fOpenIconHandle Write fOpenIconHandle;
    property ExecOnClose: Boolean Read fExecOnClose Write SetExecOnClose;
    property AutoExec: Boolean Read fAutoExec Write SetAutoExec;
    property Visible: Boolean Read fVisible Write SetVisible;
    property SortOrder: Integer Read fSortOrder Write fSortOrder;
    property Level: Integer Read fLevel Write fLevel;
    property CategoryName: String Read fCategoryName Write SetCategoryName;
    property IsMenu: Boolean Read fIsMenu Write fIsMenu;
    property MenuOpen: Boolean Read fMenuOpen Write fMenuOpen;
    property HasRun: Boolean Read fHasRun Write fHasRun;
    property ExitExe: String Read fExitExe Write SetExitExe;
    property ExitParams: String Read fExitParams Write SetExitParams;
    property IsUpdating: Boolean Read fIsUpdating Write SetIsUpdating;
    property SortString: String Read GetSortString;
    function GetPAFFile: String;
    function PAFWriteable: Boolean;
    function AppExecute: Boolean;
    function ExitExecute: Boolean;
    function ItemIndex: Integer;
    function SortKey: String;
    procedure UpdatePAFData;
    destructor Destroy; override;
  end;

  TAppItemList = class(TObject)
  private
    fAppList:          TStringList;
    fDefaultIcon:      String;  // Default Category Icon
    fDefaultOpenIcon:  String;  // Default Category Open Icon
    fReadOnly:         Boolean; // Main app must set this if drive isn't writeable
    fModified:         Boolean;
    fShowAll:          Boolean;
    fMenuIndent:       Integer; // Amount to Indent Each SubMenu.
    fDataFile:         String;  // Main Application INI File Name (Defaults to App.ini)
    fAppDBname:        String;  // DB Storage File Name (Defaults to App.fsa and Apps.fsi)
    fOnChanged:        TNotifyEvent;
    fBeforeAppExecute: TNotifyEvent;
    fAfterAppExecute:  TNotifyEvent;
    fUpdateCount:      Integer;
    fTransColor:       TColor;
    fAppFont:          TFont;
    fMenuFont:         TFont;
    fMenuOpenFont:     TFont;
    function GetAppItem(Index: Integer): TAppItem;
    procedure SetModified(bolValue: Boolean);
    procedure SetUpdateCount(intValue: Integer);
  public
    property OnChanged: TNotifyEvent Read fOnChanged Write fOnChanged;
    property BeforeAppExecute: TNotifyEvent Read fBeforeAppExecute Write fBeforeAppExecute;
    property AfterAppExecute: TNotifyEvent Read fAfterAppExecute Write fAfterAppExecute;
    property Apps[Index: Integer]: TAppItem Read GetAppItem;
    property DefaultIcon: String Read fDefaultIcon Write fDefaultIcon;
    property DefaultOpenIcon: String Read fDefaultOpenIcon Write fDefaultOpenIcon;
    property ReadOnly: Boolean Read fReadOnly Write fReadOnly;
    property Modified: Boolean Read fModified Write SetModified;
    property ShowAll: Boolean Read fShowAll Write fShowAll;
    property DataFile: String Read fDataFile Write fDataFile;
    property AppDBName: String Read fAppDBName Write fAppDBName;
    property UpdateCount: Integer Read fUPdateCount Write SetUpdateCount;
    property TransColor: TColor Read fTransColor Write fTransColor;
    property AppFont: TFont Read fAppFont Write fAppFont;
    property MenuFont: TFont Read fMenuFont Write fMenuFont;
    property MenuOpenFont: TFont Read fMenuOpenFont Write fMenuOpenFont;
    property MenuIndent: Integer Read fMenuIndent Write fMenuIndent;
    function Count: Integer;
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure Delete(ItemIndex: Integer);
    function IsVisible(ItemIndex: Integer): Boolean;
    function FirstVisible: Integer;
    function PrevVisible(From: Integer): Integer;
    function NextVisible(From: Integer): Integer;
    function LastVisible: Integer;
    function AtTop(ItemIndex: Integer): Boolean;
    function AtBottom(ItemIndex: Integer; DisplayCount: Integer): Boolean;
    function AddApp(ExePath: String): Integer; overload;
    function AddApp(DBID: Integer; DB: TIniFile; IconStream: TStream): Integer; overload;
    function AddCategory(strName: String; strCaption: String; strCategory: String;
      bolIsOpen: Boolean; intSortOrder: Integer; strIcoPath: String;
      strOpenIcoPath: String): Integer;
    procedure DelCategory(strName: String); overload;
    procedure DelCategory(ItemIndex: Integer); overload;
    function GetCtgByName(CtgName: String): Integer;
    procedure ToggleMenu(ItemIndex: Integer);
    procedure SaveToDB; overload;
    function LoadFromDB: Boolean; overload;
    procedure SaveToDB(DBFileName: String); overload;
    function LoadFromDB(DBFileName: String): Boolean; overload;
    function LoadFromPath(strPortableAppsPath: String; Excludes: TStrings): Boolean;
    function LoadCategories: Integer;
    procedure SaveCategories;
    procedure ChangeCtgIcon(NewIcon: String);
    procedure ChangeOpenCtgIcon(NewIcon: String);
    procedure CategoryMenu(ItemIndex: Integer; mnuParent: TMenuItem);
    procedure mnuCategoryClick(Sender: TObject);
    procedure Apply(intIndex: Integer; sbtnTarget: TSpeedButton);
    procedure DrawButtons(var Buttons: array of TSpeedButton);
    procedure ShowSubItems(ItemIndex: Integer; var Buttons: array of TSpeedButton);
    procedure FixLevels; // Really bad hack, should traverse tree rather then doing all
    function CanSortUp(ItemIndex: Integer): Boolean;
    function CanSortDown(ItemIndex: Integer): Boolean;
    procedure SortUp(ItemIndex: Integer);
    procedure SortDown(ItemIndex: Integer);
    procedure SortTop(ItemIndex: Integer);
    procedure SortBottom(ItemIndex: Integer);
    function PrepareSort(ItemIndex: Integer; slWork: TStringList): Integer;
    procedure FinalizeSort(slWork: TStringList);
    procedure Scroll(var Buttons: array of TSpeedButton; intSteps: Integer);
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure AutoExec;
    function AppExecute(ItemIndex: Integer): Boolean;
    procedure SetFontSize(intSize: Integer);
    procedure ExitExec;
  end;

const
  AppDBVersion = '1.0';


implementation

function StripDrive(Path: String): String;
begin
  Result := Path;
  if (Length(Path) > 2) then
  begin
    if (Path[2] = ':') then
    begin
      Result := Copy(Path, 3, Length(Path) - 2);
    end;
  end;
end;

constructor TAppItem.Create(intDBID: Integer; DB: TIniFile; IconStream: TStream);
var
  Section: String;
  Istr:    TMemoryStream;
  Isz:     Integer;
  Ico:     TIcon;
  Drive:   String;
begin
  inherited Create;
  Drive         := ExtractFileDrive(Application.ExeName);
  Section       := format('APP-%3.3d', [intDBID]);
  fName         := DB.ReadString(Section, 'Name', '');
  fCaption      := DB.ReadString(Section, 'Caption', fName);
  fExePath      := Drive + StripDrive(DB.ReadString(Section, 'ExePath', ''));
  fHint         := DB.ReadString(Section, 'Hint', 'Location: ' + ExtractFileDir(fExePath));
  fExecOnClose  := DB.ReadBool(Section, 'ExecOnClose', False);
  fAutoExec     := DB.ReadBool(Section, 'AutoExec', False);
  fVisible      := DB.ReadBool(Section, 'Visible', True);
  fSortOrder    := DB.ReadInteger(Section, 'SortOrder', 0);
  fLevel        := DB.ReadInteger(Section, 'Level', 0);
  fCategoryName := DB.ReadString(Section, 'CategoryName', '');
  fIsMenu       := DB.ReadBool(Section, 'IsMenu', False);
  fMenuOpen     := DB.ReadBool(Section, 'MenuOpen', False);
  fExitExe      := DB.ReadString(Section, 'ExitExe', '');
  if fExitExe <> '' then
  begin
    fExitExe := Drive + StripDrive(fExitExe);
  end;
  fExitParams := DB.ReadString(Section, 'ExitParams', '');
  fIsUpdating := False;
  fHasRun := False;
  fIconHandle := 0;
  fOpenIconHandle := 0;
  Isz := DB.ReadInteger(Section, 'IcoSize', 0);
  if (Isz > 0) then
  begin
    Ico  := TIcon.Create;
    IStr := TMemoryStream.Create;
    Istr.CopyFrom(IconStream, Isz);
    Istr.Seek(0, soFromBeginning);
    Ico.LoadFromStream(Istr);
    Istr.Clear;
    Istr.Free;
    fIconHandle := CopyIcon(Ico.Handle);
    Ico.Free;
  end;
  if fIsMenu then
  begin
    Isz := DB.ReadInteger(Section, 'OpenIcoSize', 0);
    if (Isz > 0) then
    begin
      Ico  := TIcon.Create;
      IStr := TMemoryStream.Create;
      Istr.CopyFrom(IconStream, Isz);
      Istr.Seek(0, soFromBeginning);
      Ico.LoadFromStream(Istr);
      Istr.Clear;
      Istr.Free;
      fOpenIconHandle := CopyIcon(Ico.Handle);
      Ico.Free;
    end;
  end;
end;

constructor TAppItem.Create(ExePath: String);
var
  ptrFileInfo: pointer;
  lngFileInfoSize: Longint;
  dwdFileInfoLength: DWORD;
  ptrFileInfoLanguage: Pointer;
  ptrFileInfoData: Pointer;
  strFileInfoQuery: String;
  uintFileInfoDataLength: Uint;
  FileInfo: TSHFileInfo;
begin
  inherited Create;
  fName           := '';
  fCaption        := '';
  fExePath        := ExePath;
  fHint           := '';
  fIconHandle     := 0;
  fOpenIconHandle := 0;
  fExecOnClose    := False;
  fAutoExec       := False;
  fVisible        := True; // is Application Hidden
  fSortOrder      := 0;
  fLevel          := 0;
  fCategoryName   := '';
  fIsMenu         := False;
  fMenuOpen       := False;
  fExitExe        := '';
  fExitParams     := '';
  fHasRun         := False;
  fIsUpdating     := False;
  if FileExists(fExePath) then
  begin
    lngFileInfoSize := GetFileVersionInfoSize(PChar(fExePath), dwdFileInfoLength);
    if (lngFileInfoSize > 0) then
    begin
      GetMem(ptrFileInfo, lngFileInfoSize);
      try
        // Check if the current app has a ProductName
        if GetFileVersionInfo(PChar(fExePath), dwdFileInfoLength, lngFileInfoSize,
          ptrFileInfo) then
        begin
          strFileInfoQuery := 'ProductName';
          if VerQueryValue(ptrFileInfo, '\VarFileInfo\Translation',
            ptrFileInfoLanguage, uintFileInfoDataLength) then
          begin
            strFileInfoQuery := Format('\StringFileInfo\%0.4x%0.4x\%s'#0,
              [LoWord(Longint(ptrFileInfoLanguage^)), HiWord(Longint(ptrFileInfoLanguage^)),
              strFileInfoQuery]);
          end;
          if VerQueryValue(ptrFileInfo, PChar(strFileInfoQuery),
            ptrFileInfoData, uintFileInfoDataLength) then
          begin
            fName := strPas(ptrFileInfoData);
          end;
        end;
      finally
        FreeMem(ptrFileInfo, lngFileInfoSize);
      end;
      GetMem(ptrFileInfo, lngFileInfoSize);
      if (trim(fName) = '') then
      begin
        try
          // Check if the current app has a FileDescription
          if GetFileVersionInfo(PChar(fExePath), dwdFileInfoLength,
            lngFileInfoSize, ptrFileInfo) then
          begin
            strFileInfoQuery := 'FileDescription';
            if VerQueryValue(ptrFileInfo, '\VarFileInfo\Translation',
              ptrFileInfoLanguage, uintFileInfoDataLength) then
            begin
              strFileInfoQuery := Format('\StringFileInfo\%0.4x%0.4x\%s'#0,
                [LoWord(Longint(ptrFileInfoLanguage^)), HiWord(Longint(ptrFileInfoLanguage^)),
                strFileInfoQuery]);
            end;
            if VerQueryValue(ptrFileInfo, PChar(strFileInfoQuery),
              ptrFileInfoData, uintFileInfoDataLength) then
            begin
              fName := strPas(ptrFileInfoData);
            end;
          end;
        finally
          FreeMem(ptrFileInfo, lngFileInfoSize);
        end;
      end;
    end;
    if (trim(fName) = '') then
    begin
      fName := ChangeFileExt(ExtractFileName(fExePath), '');
    end; // Default Name
    if FileExists(GetPAFFile) then
    begin
      with TIniFile.Create(GetPAFFile) do
      begin
        fCaption      := ReadString('PAFDATA', 'Name', fName);
        fVisible      := StrToBoolDef(ReadString('PAFDATA', 'Visible', 'True'), True);
        fAutoExec     := StrToBoolDef(ReadString('PAFDATA', 'AutoExec', 'False'), False);
        fHint         := ReadString('PAFDATA', 'Hint', fHint);
        fExecOnCLose  := StrToBoolDef(ReadString('PAFDATA', 'ExecOnClose', 'False'), False);
        // Moved to app.ini        fSortOrder := StrToIntDef(ReadString('PAFDATA', 'SortHint', '0'), 0);
        // AppItemList handles loading.  (Not the best, should be fixed)
        fCategoryName := ReadString('PAFDATA', 'Category', fCategoryName);
        fExitExe      := ReadString('PAFDATA', 'ExitExe', fExitExe);
        fExitParams   := ReadString('PAFDATA', 'ExitParams', fExitParams);
        Free;
      end;
    end;
    if (fCaption = '') then
    begin
      fCaption := fName;
    end;
    if (fHint = '') then
    begin
      fHint := 'Location: ' + ExtractFileDir(fExePath);
    end;
    FillChar(FileInfo, SizeOf(FileInfo), #0);
    SHGetFileInfo(PChar(fExePath), 0, FileInfo, SizeOf(FileInfo), SHGFI_ICON or
      SHGFI_SMALLICON);
    fIconHandle := FileInfo.hIcon;
  end;
end;

destructor TAppItem.Destroy;
begin
  if (fIconHandle <> 0) then
  begin
    DestroyIcon(fIconHandle);
  end;
  if (fOpenIconHandle <> 0) then
  begin
    DestroyIcon(fOpenIconHandle);
  end;
  inherited Destroy;
end;

procedure TAppItem.AddOpenIcon(strIcoPath: String);
var
  FileInfo: TSHFileInfo;
begin
  if (fOpenIconHandle <> 0) then
  begin
    DestroyIcon(fOpenIconHandle);
    fOpenIconHandle := 0;
    fExitExe        := '';
  end;
  if Assigned(fAppList) then
  begin
    if (strIcoPath = '') and (TAppItemList(fAppList).DefaultOpenIcon <> '') then
    begin
      strIcoPath := TAppItemList(fAppList).DefaultOpenIcon;
    end;
    if (strIcoPath = TAppItemList(fAppList).DefaultIcon) or (strIcoPath = '') then
    begin
      Exit; // Is the same as the default or blank so no need to load
    end;
    if (strIcoPath <> TAppItemList(fAppList).DefaultOpenIcon) then
    begin
      fExitExe := strIcoPath;
    end;
  end;
  if FileExists(strIcoPath) then
  begin
    FillChar(FileInfo, SizeOf(FileInfo), #0);
    SHGetFileInfo(PChar(strIcoPath), 0, FileInfo, SizeOf(FileInfo), SHGFI_ICON or
      SHGFI_SMALLICON);
    fOpenIconHandle := FileInfo.hIcon;
  end;
end;

procedure TAppItem.SetModified;
begin
  if Assigned(fAppList) then
  begin
    TAppItemList(fAppList).Modified := True;
  end;
end;

procedure TAppItem.UpdateSort;
begin
  if Assigned(fAppList) then
  begin
    TAppItemList(fAppList).fAppList.Sorted     := False;
    TAppItemList(fAppList).fAppList[ItemIndex] := SortString;
    TAppItemList(fAppList).fAppList.Sort;
    TAppItemList(fAppList).fAppList.Sorted     := True;
    TAppItemList(fAppList).FixLevels;
  end;
end;

procedure TAppItem.SetCaption(strValue: String);
begin
  if strValue <> fCaption then
  begin
    fCaption := strValue;
    UpdatePAFData;
    UpdateSort;
    SetModified;
  end;
end;

procedure TAppItem.SetHint(strValue: String);
begin
  if strValue <> fHint then
  begin
    fHint := strValue;
    UpdatePAFData;
    SetModified;
  end;
end;

procedure TAppItem.SetExecOnClose(bolValue: Boolean);
begin
  if bolValue <> fExecOnClose then
  begin
    fExecOnClose := bolValue;
    UpdatePAFData;
    SetModified;
  end;
end;

procedure TAppItem.SetAutoExec(bolValue: Boolean);
begin
  if bolValue <> fAutoExec then
  begin
    fAutoExec := bolValue;
    UpdatePAFData;
    SetModified;
  end;
end;

procedure TAppItem.SetVisible(bolValue: Boolean);
begin
  if bolValue <> fVisible then
  begin
    fVisible := bolValue;
    UpdatePAFData;
    SetModified;
  end;
end;

procedure TAppItem.SetCategoryName(strValue: String);
begin
  if strValue <> fCategoryName then
  begin
    fCategoryName := strValue;
    UpdatePAFData;
    UpdateSort;
    if Assigned(fAppList) then
    begin
      TAppItemList(fAppList).FixLevels;
    end;
    SetModified;
  end;
end;

procedure TAppItem.SetExitExe(strValue: String);
begin
  if strValue <> fExitExe then
  begin
    fExitExe := strValue;
    UpdatePAFData;
    SetModified;
  end;
end;

procedure TAppItem.SetExitParams(strValue: String);
begin
  if strValue <> fExitParams then
  begin
    fExitParams := strValue;
    UpdatePAFData;
    SetModified;
  end;
end;

procedure TAppItem.SetIsUpdating(bolValue: Boolean);
begin
  if bolValue <> fIsUpdating then
  begin
    fIsUpdating := bolValue;
    if not fIsUpdating then
    begin
      UpdatePAFData;
    end;
  end;
end;

procedure TAppItem.UpdatePAFData;
var
  bolDefault: Boolean;  // Track if any of the items are non default
begin
  if fIsUpdating then
  begin
    Exit;
  end; // Multiple Changes so wait till IsUpdating is Unset
  if fIsMenu then
  begin
        //  if Assigned(fAppList) then
        //    TAppItemList(fAppList).SaveCategories;
  end else
  begin // Standard App
    if FileExists(GetPAFFile) and FileIsReadOnly(GetPAFFile) then
    begin
      Exit;
    end; // File is readonly
    if (Assigned(fAppList)) then
    begin
      if TAppItemList(fAppList).ReadOnly then
      begin
        Exit;
      end;
    end; // Main app has specified read only
    bolDefault := True;
    with TIniFile.Create(GetPAFFile) do
    begin
      try
        if (fCaption <> fName) then
        begin
          WriteString('PAFDATA', 'Name', fCaption);
          bolDefault := False;
        end else
        begin
          DeleteKey('PAFDATA', 'Name');
        end;
        if (not fVisible) then
        begin
          WriteString('PAFDATA', 'Visible', BoolToStr(fVisible, True));
          bolDefault := False;
        end else
        begin
          DeleteKey('PAFDATA', 'Visible');
        end;
        if (fAutoExec) then
        begin
          WriteString('PAFDATA', 'AutoExec', BoolToStr(fAutoExec, True));
          bolDefault := False;
        end else
        begin
          DeleteKey('PAFDATA', 'AutoExec');
        end;
        if (fHint <> 'Location: ' + ExtractFileDir(fExePath)) then
        begin
          WriteString('PAFDATA', 'Hint', fHint);
          bolDefault := False;
        end else
        begin
          DeleteKey('PAFDATA', 'Hint');
        end;
        if (fExecOnClose) then
        begin
          WriteString('PAFDATA', 'ExecOnClose', BoolToStr(fExecOnClose, True));
          bolDefault := False;
        end else
        begin
          DeleteKey('PAFDATA', 'ExecOnClose');
        end;
{ Moved Sort Order into App.ini otherwise changing the sort
  could result in every single PAF file on the disk being written
    if (fSortOrder <> 0) then
    begin
      WriteInteger('PAFDATA', 'SortOrder', fSortOrder);
      bolDefault := False;
    end else begin
      DeleteKey('PAFDATA', 'SortOrder');
    end; }
        if (fCategoryName <> '') then
        begin
          WriteString('PAFDATA', 'Category', fCategoryName);
          bolDefault := False;
        end else
        begin
          DeleteKey('PAFDATA', 'Category');
        end;
        if (fExitExe <> '') then
        begin
          WriteString('PAFDATA', 'ExitExe', StripDrive(fExitExe));
          bolDefault := False;
        end else
        begin
          DeleteKey('PAFDATA', 'ExitExe');
        end;
        if (fExitParams <> '') then
        begin
          WriteString('PAFDATA', 'ExitParams', fExitParams);
          bolDefault := False;
        end else
        begin
          DeleteKey('PAFDATA', 'ExitParams');
        end;
      finally
        Free;
      end;
    end;
    if FileExists(GetPAFFile) and bolDefault then
    begin
      DeleteFile(GetPAFFile);
    end; // All Values are at default so kill PAF file
  end;
end;

function TAppItem.AppExecute: Boolean;
var
  strDir: String;
begin
  Result := False;
  if Assigned(fAppList) then
  begin
    if Assigned(TAppItemList(fAppList).BeforeAppExecute) then
    begin
      TAppItemList(fAppList).BeforeAppExecute(Self);
    end;
  end;
  if isMenu then
  begin
    if Assigned(fAppList) then
    begin
      TAppItemList(fAppList).ToggleMenu(ItemIndex);
    end;
  end else if FileExists(fExePath) then
  begin
    strDir  := ExtractFilePath(fExePath);
    fHasRun := (ShellExecute(0, 'open', PAnsiChar(fExePath), nil, PAnsiChar(strDir),
      SW_SHOW) > 32);
    Result  := fHasRun;
  end;
  if Assigned(fAppList) then
  begin
    if Assigned(TAppItemList(fAppList).AfterAppExecute) then
    begin
      TAppItemList(fAppList).AfterAppExecute(Self);
    end;
  end;
end;

function TAppItem.ExitExecute: Boolean;
var
  strExe: String;
  strDir: String;
  strPrm: String;
begin
  Result := False;
  strExe := fExePath;
  strPrm := fExitParams;
  if fExitExe <> '' then
  begin
    // if ExeExe is set the only do exitexec if app has run
    if fHasRun then
    begin
      strExe := fExitExe;
    end
    else
    begin
      strExe := '';
    end;
  end;
  if strExe <> '' then
  begin
    if FileExists(strExe) then
    begin
      strDir := ExtractFilePath(strExe);
      ShellExecute(0, 'open', PAnsiChar(strExe), PAnsiChar(strPrm), PAnsiChar(strDir), SW_SHOW);
      Result := fHasRun;
    end;
  end;
end;

function TAppItem.GetSortString: String;
var
  strWork: String;
  intCtg:  Integer;
begin
  strWork := Format('%3.3d %-20.20s', [fSortOrder, fCaption]);
  if (fCategoryName <> '') then
  begin
    if Assigned(fAppList) then
    begin
      with TAppItemList(fAppList) do
      begin
        intCtg := GetCtgByName(fCategoryName);
        if (intCtg >= 0) then
        begin
          strWork := Apps[intCtg].GetSortString + strWork;
        end;
      end;
    end;
  end;
  Result := strWork;
end;

procedure TAppItem.SaveToDB(DBID: Integer; DB: TIniFile; IconStream: tStream);
var
  Section: String;
  Istr:    TMemoryStream;
  Ico:     TIcon;
begin
  Section := format('APP-%3.3d', [DBID]);
  DB.WriteString(Section, 'Name', fName);
  if (fCaption <> fName) then
  begin
    DB.WriteString(Section, 'Caption', fCaption);
  end;
  DB.WriteString(Section, 'ExePath', StripDrive(fExePath));
  if (Hint <> 'Location: ' + ExtractFileDir(fExePath)) then
  begin
    DB.WriteString(Section, 'Hint', fHint);
  end;
  if (fExecOnClose) then
  begin
    DB.WriteBool(Section, 'ExecOnClose', fExecOnClose);
  end;
  if (fAutoExec) then
  begin
    DB.WriteBool(Section, 'AutoExec', fAutoExec);
  end;
  if (not fVisible) then
  begin
    DB.WriteBool(Section, 'Visible', fVisible);
  end;
  if (fSortOrder <> 0) then
  begin
    DB.WriteInteger(Section, 'SortOrder', fSortOrder);
  end;
  if (fLevel <> 0) then
  begin
    DB.WriteInteger(Section, 'Level', fLevel);
  end;
  if (fCategoryName <> '') then
  begin
    DB.WriteString(Section, 'CategoryName', fCategoryName);
  end;
  if (fIsMenu) then
  begin
    DB.WriteBool(Section, 'IsMenu', fIsMenu);
  end;
  if (fMenuOpen) then
  begin
    DB.WriteBool(Section, 'MenuOpen', fMenuOpen);
  end;
  if (fExitExe <> '') then
  begin
    DB.WriteString(Section, 'ExitExe', StripDrive(fExitExe));
  end;
  if (fExitParams <> '') then
  begin
    DB.WriteString(Section, 'ExitParams', fExitParams);
  end;
  if (IconHandle <> 0) then
  begin
    Ico        := TIcon.Create;
    Ico.Handle := IconHandle;
    Istr       := TMemoryStream.Create;
    Istr.Clear;
    Ico.SaveToStream(Istr);
    DB.WriteInteger(Section, 'IcoSize', Istr.Size);
    Istr.Seek(0, soFromBeginning);
    IconStream.CopyFrom(Istr, Istr.Size);
    Istr.Clear;
    Istr.Free;
    Ico.ReleaseHandle;
    Ico.Free;
  end;
  if (OpenIconHandle <> 0) then
  begin
    Ico        := TIcon.Create;
    Ico.Handle := OpenIconHandle;
    Istr       := TMemoryStream.Create;
    Istr.Clear;
    Ico.SaveToStream(Istr);
    DB.WriteInteger(Section, 'OpenIcoSize', Istr.Size);
    Istr.Seek(0, soFromBeginning);
    IconStream.CopyFrom(Istr, Istr.Size);
    Istr.Clear;
    Istr.Free;
    Ico.ReleaseHandle;
    Ico.Free;
  end;
end;

function TAppItem.GetPAFFile: String;
begin
  Result := ChangeFileExt(fExePath, '.pafdata');
end;

function TAppItem.PAFWriteable: Boolean;
begin
  Result := True;
  if FileExists(GetPAFFile) then
  begin
    if FileIsReadOnly(GetPAFFile) then
    begin
      Result := False;
    end;
  end;
end;

function TAppItem.ItemIndex: Integer;
var
  I: Integer;
begin
  Result := -1;
  if Assigned(fAppList) then
  begin
    for I := 0 to TAppItemList(fAppList).Count - 1 do
    begin
      if TAppItemList(fAppList).fAppList.Objects[I] = Self then
      begin
        Result := I;
        Break;
      end;
    end;
  end;
end;

function TAppItem.SortKey: String;
  // String to Use for INI Files SortOrder Key
begin
  if fIsMenu then
  begin
    Result := 'Category-' + StringReplace(fName, '=', '|eq|', [rfReplaceAll, rfIgnoreCase]);
  end
  else
  begin
    Result := StringReplace(StripDrive(fExePath), '=', '|eq|', [rfReplaceAll, rfIgnoreCase]);
  end;
end;



constructor TAppItemList.Create;
begin
  inherited Create;
  fDefaultIcon       := ExtractFilePath(Application.ExeName);
  fDefaultOpenIcon   := '';
  fDataFile          := ChangeFileExt(Application.ExeName, '.ini');
  fAppDBname         := ChangeFileExt(Application.ExeName, '.fsa');
  fShowAll           := False;
  fModified          := False;
  fAppList           := TStringList.Create;
  fAppList.Sorted    := True;
  fAppList.Duplicates := dupAccept;
  fReadOnly          := False;
  fMenuIndent        := 10;
  fUpdateCount       := 0;
  fAppFont           := TFont.Create;
  fAppFont.Color     := clBlack;
  fAppFont.Size      := 8;
  fAppFont.Name      := 'Tahoma';
  fAppFont.Style     := [fsBold];
  fMenuFont          := TFont.Create;
  fMenuFont.Color    := clBlack;
  fMenuFont.Size     := 8;
  fMenuFont.Name     := 'Tahoma';
  fMenuFont.Style    := [fsBold];
  fMenuOpenFont      := TFont.Create;
  fMenuOpenFont.Color := clBlack;
  fMenuOpenFont.Size := 8;
  fMenuOpenFont.Name := 'Tahoma';
  fMenuOpenFont.Style := [fsBold];
end;

destructor TAppItemList.Destroy;
begin
  Clear;
  fAppList.Free;
  fMenuOpenFont.Free;
  fMenuFont.Free;
  fAppFont.Free;
  //  fCategories.Free;
  inherited Destroy;
end;

procedure TAppItemList.Clear;
begin
  while fAppList.Count > 0 do
  begin
    fAppList.Objects[0].Free;
    fAppList.Delete(0);
  end;
  fAppList.Clear;
  //    fCategories.Clear;
end;

procedure TAppItemList.Delete(ItemIndex: Integer);
begin
  if (ItemIndex >= 0) and (ItemIndex < Count) then
  begin
    fAppList.Objects[ItemIndex].Free;
    fAppList.Delete(ItemIndex);
  end;
end;

procedure TAppItemList.SetModified(bolValue: Boolean);
begin
  if bolValue <> fModified then
  begin
    FixLevels;
    fModified := bolValue;
    if fModified then
    begin
      if fUpdateCount = 0 then
      begin
        if Assigned(fOnChanged) then
        begin
          fOnChanged(Self);
        end;
      end;
    end;
  end;
end;

procedure TAppItemList.SetUpdateCount(intValue: Integer);
begin
  if fUpdateCount <> intValue then
  begin
    if intValue > 0 then
    begin
      fUpdateCount := intValue;
    end
    else
    begin
      fUpdateCount := 0;
    end;
  end;
  if fUpdateCount = 0 then
  begin
    if Assigned(fOnChanged) then
    begin
      fOnChanged(Self);
    end;
  end;
end;

function TAppItemList.Count: Integer;
begin
  Result := fAppList.Count;
end;

function TAppItemList.IsVisible(ItemIndex: Integer): Boolean;
var
  I: Integer;
  L: Integer;
begin
  Result := False;
  if (ItemIndex < 0) or (ItemIndex >= fAppList.Count) then
  begin
    Exit;
  end;
  if not Apps[ItemIndex].Visible and not ShowAll then
  begin
    Exit;
  end;
  if Apps[ItemIndex].Level = 0 then
  begin
    Result := True;
    Exit;
  end;
  L := Apps[ItemIndex].Level;
  I := ItemIndex;
  while L > 0 do
  begin
    while Apps[I].Level >= L do
    begin
      Dec(I);
    end;
    if (not Apps[I].Visible and not ShowAll) or (not Apps[I].MenuOpen) then
    begin
      Exit;
    end; // Menu is closed or hidden then app in menu is not visible
    L := Apps[I].Level; // Set L to check higher level
  end;
  Result := True;
end;

function TAppItemList.FirstVisible: Integer;
var
  I: Integer;
begin
  Result := -1;
  if Count > 0 then
  begin
    for I := 0 to fAppList.Count - 1 do
    begin
      if IsVisible(I) then
      begin
        Result := I;
        Break;
      end;
    end;
  end;
end;

function TAppItemList.PrevVisible(From: Integer): Integer;
var
  I: Integer;
begin
  Result := -1;
  if From >= Count then
  begin
    From := Count - 1;
  end; // Max val coming in has to be count-1
  if From > 0 then
  begin
    for I := From - 1 downto 0 do
    begin
      if IsVisible(I) then
      begin
        Result := I;
        Break;
      end;
    end;
  end;
end;

function TAppItemList.NextVisible(From: Integer): Integer;
var
  I: Integer;
begin
  Result := -1;
  if From < -1 then
  begin
    From := 0;
  end;
  if From < Count - 1 then
  begin
    for I := From + 1 to Count - 1 do
    begin
      if IsVisible(I) then
      begin
        Result := I;
        Break;
      end;
    end;
  end;
end;

function TAppItemList.LastVisible: Integer;
var
  I: Integer;
begin
  Result := -1;
  if Count > 0 then
  begin
    for I := fAppList.Count - 1 downto 0 do
    begin
      if IsVisible(I) then
      begin
        Result := I;
        Break;
      end;
    end;
  end;
end;

function TAppItemList.AtTop(ItemIndex: Integer): Boolean;
begin
  Result := (FirstVisible >= ItemIndex);
end;

function TAppItemList.AtBottom(ItemIndex: Integer; DisplayCount: Integer): Boolean;
var
  I: Integer;
  C: Integer;
begin
  I := LastVisible;
  C := 1;
  while (C < DisplayCount) and (I > ItemIndex) do
  begin
    I := PrevVisible(I);
    Inc(C);
  end;
  Result := (C < DisplayCount) or (I <= ItemIndex);
end;

function TAppItemList.GetAppItem(Index: Integer): TAppItem;
begin
  Result := TAppItem(fAppList.Objects[Index]);
end;

function TAppItemList.GetCtgByName(CtgName: String): Integer;
var
  I: Integer;
begin
  Result := -1;
  if CtgName = '' then
  begin
    Exit;
  end;
  if Count > 0 then
  begin
    for I := 0 to Count - 1 do
    begin
      if TAppItem(fAppList.Objects[I]).IsMenu then
      begin
        if (TAppItem(fAppList.Objects[I]).fName = CtgName) then
        begin
          Result := I;
          Break;
        end;
      end;
    end;
  end;
end;

procedure TAppItemList.ToggleMenu(ItemIndex: Integer);
begin
  if (ItemIndex >= 0) and (ItemIndex < Count) then
  begin
    if Apps[ItemIndex].IsMenu then
    begin
      BeginUpdate;
      Apps[ItemIndex].MenuOpen := not Apps[ItemIndex].MenuOpen;
      EndUpdate;
    end;
  end;
end;

function TAppItemList.AddApp(ExePath: String): Integer;
var
  App:    TAppItem;
  intCtg: Integer;
begin
  App    := TAppItem.Create(ExePath);
  Result := -1;
  if Assigned(App) then
  begin
    if (App.Name = 'OpenOffice.org Base Portable') then
    begin
      // Kludge to force ignoring Openoffice Base Program
      App.Free;
    end else
    begin
      App.fAppList := Self;
      intCtg       := GetCtgByName(App.CategoryName);
      if (intCtg >= 0) then
      begin
        App.MenuOpen := Apps[intCtg].MenuOpen;
        App.Level    := Apps[intCtg].Level + 1;
      end;
      with TIniFile.Create(DataFile) do
      begin
        App.SortOrder := ReadInteger('SortOrder', App.SortKey, 0);
        Free;
      end;
      Result := fAppList.AddObject(App.SortString, App);
    end;
  end;
  fModified := True;
end;


function TAppItemList.AddApp(DBID: Integer; DB: TIniFile; IconStream: TStream): Integer;
var
  App:    TAppItem;
  intCtg: Integer;
begin
  App := TAppItem.Create(DBID, DB, IconStream);
  if Assigned(App) then
  begin
    App.fAppList := Self;
    intCtg       := GetCtgByName(App.CategoryName);
    if (intCtg >= 0) then
    begin
      App.MenuOpen := Apps[intCtg].MenuOpen;
      App.Level    := Apps[intCtg].Level + 1;
    end;
    Result := fAppList.AddObject(App.SortString, App);
  end else
  begin
    Result := -1;
  end;
end;

function TAppItemList.AddCategory(strName: String; strCaption: String;
  strCategory: String; bolIsOpen: Boolean; intSortOrder: Integer; strIcoPath: String;
  strOpenIcoPath: String): Integer;
var
  App:    TAppItem;
  intCtg: Integer;
begin
  if strIcoPath = '' then
  begin
    strIcoPath := DefaultIcon;
  end;
  Result := -1;
  intCtg := GetCtgByName(strName);
  if intCtg = -1 then
  begin
    App := TAppItem.Create(strIcoPath);
    if (strIcoPath = DefaultIcon) then
    begin
      App.fExePath := '';
    end;
    if Assigned(App) then
    begin
      App.IsUpdating   := True;
      App.Name         := strName;
      App.Caption      := strCaption;
      App.Hint         := '';
      App.IsMenu       := True;
      App.CategoryName := strCategory;
      App.MenuOpen     := bolIsOpen;
      App.SortOrder    := intSortOrder;
      App.fAppList     := Self;
      intCtg           := GetCtgByName(App.CategoryName);
      if (intCtg >= 0) then
      begin
        App.MenuOpen := Apps[intCtg].MenuOpen;
        App.Level    := Apps[intCtg].Level + 1;
      end;
      App.AddOpenIcon(strOpenIcoPath); // add open icon (if any)
      App.IsUpdating := False;
      fModified      := True;
      with TIniFile.Create(DataFile) do
      begin
        App.SortOrder := ReadInteger('SortOrder', App.SortKey, 0);
        Free;
      end;
      Result := fAppList.AddObject(App.SortString, App);
      FixLevels;
    end;
  end;
end;

procedure TAppItemList.DelCategory(strName: String);
begin
  DelCategory(GetCtgByName(strName));
end;

procedure TAppItemList.DelCategory(ItemIndex: Integer);
var
  I: Integer;
begin
  if (ItemIndex >= 0) and (ItemIndex < Count) then
  begin
    if Apps[ItemIndex].IsMenu then
    begin
      BeginUpdate;
      I := ItemIndex + 1;
      while (I < Count) do
      begin
        if (Apps[I].Level > Apps[ItemIndex].Level) then
        begin
          Apps[I].Level := Apps[I].Level - 1;
          Inc(I);
        end else
        begin
          I := Count;
        end;
      end;
      fAppList.Objects[ItemIndex].Free;
      fAppList.Delete(ItemIndex);
      FixLevels;
      EndUpdate;
    end;
  end;
end;

procedure TAppItemList.SaveToDB;
begin
  SaveToDB(fAppDBname);
end;

procedure TAppItemList.SaveToDB(DBFileName: String);
var
  DB: TIniFile;
  IconFileName: String;
  IconStream: TFileStream;
  I: Integer;
begin
  if FileExists(DBFileName) then
  begin
    DeleteFile(DBFileName);
  end;
  IconFileName := ChangeFileExt(DBFileName, '.fsi');
  if FileExists(IconFileName) then
  begin
    DeleteFile(IconFileName);
  end;
  if (fAppList.Count > 0) then
  begin
    DB := TIniFile.Create(DBFileName);
    DB.WriteString('DB', 'Version', AppDBVersion);
    DB.WriteInteger('DB', 'Count', fAppList.Count);
    IconStream := TFileStream.Create(IconFileName, fmCreate);
    for I := 0 to fAppList.Count - 1 do
    begin
      Apps[I].SaveToDB(I, DB, IconStream);
    end;
    DB.WriteInteger('DB', 'TotalIconSize', IconStream.Size);
    IconStream.Free;
    DB.Free;
  end;
end;

function TAppItemList.LoadFromDB: Boolean;
begin
  Result := LoadFromDB(fAppDBname);
end;

function TAppItemList.LoadFromDB(DBFileName: String): Boolean;
var
  DB: TIniFile;
  IconFileName: String;
  IconStream: TFileStream;
  I: Integer;
  C: Integer;
  S: String;
begin
  Result := False;
  Clear;
  if FileExists(DBFileName) then
  begin
    IconFileName := ChangeFileExt(DBFileName, '.fsi');
    if FileExists(IconFileName) then
    begin
      DB := TIniFile.Create(DBFileName);
      try // Used so an Exit will close the file properly
        S := DB.ReadString('DB', 'Version', '');
          // Minor Version Changes are OK
        if (Copy(S, 0, Pos('.', S)) <> Copy(AppDBVersion, 0, Pos('.', AppDBVersion))) then
        begin
          Exit;
        end;
        IconStream := TFileStream.Create(IconFileName, fmOpenRead);
        try
          I := IconStream.Size;
          // Fail if IconFile Size is wrong
          if (DB.ReadInteger('DB', 'TotalIconSize', 0) <> I) then
          begin
            Exit;
          end;
          C := DB.ReadInteger('DB', 'Count', 0);
          if (C > 0) then
          begin
            for I := 0 to C - 1 do
            begin
              AddApp(I, DB, IconStream);
            end;
          end;
          Result := True;
        finally
          IconStream.Free;
        end;
      finally
        DB.Free;
      end;
    end;
  end;
end;

function TAppItemList.LoadFromPath(strPortableAppsPath: String; Excludes: TStrings): Boolean;
var
  tsrchDirectory:    TSearchRec;
  tsrchSubDirectory: TSearchRec;
  strTempSubDir:     String;
begin
  Result := False;
  // Always ignore current and parent directories
  Excludes.Add('.');
  Excludes.Add('..');
  if FindFirst(IncludeTrailingPathDelimiter(strPortableAppsPath) + '*',
    faDirectory, tsrchDirectory) = 0 then
  begin
    repeat
      strTempSubDir := tsrchDirectory.Name;
      // Filter Out Excluded Directories
      if (Excludes.IndexOf(strTempSubDir) = -1) then
      begin
        strTempSubDir := IncludeTrailingPathDelimiter(
          IncludeTrailingPathDelimiter(strPortableAppsPath) + strTempSubDir);
        if FindFirst(strTempSubDir + '*.exe', faAnyFile, tsrchSubDirectory) = 0 then
        begin
          repeat
            // Ensure it's not a directory with a .exe name
            if ((tsrchSubDirectory.Attr and faDirectory) = 0) then
            begin
              AddApp(strTempSubDir + tsrchSubDirectory.Name);
              Result := True;
            end;
          until FindNext(tsrchSubDirectory) <> 0;
          FindClose(tsrchSubDirectory);
        end;
      end;
    until FindNext(tsrchDirectory) <> 0;
    FindClose(tsrchDirectory);
  end;
end;

function TAppItemList.LoadCategories: Integer;
var
  I:           Integer;
  intNew:      Integer;
  strSection:  String;
  strIcon:     String;
  strOpenIcon: String;
  strName:     String;
  strCaption:  String;
  strCategory: String;
  strDrive:    String;
begin
  Result   := 0;
  strDrive := ExtractFileDrive(Application.ExeName);
  with TIniFile.Create(DataFile) do
  begin
    I          := 0;
    strSection := Format('Ctg_%d', [I]);
    while SectionExists(strSection) do
    begin
      strName     := ReadString(strSection, 'Name', strSection);
      strCaption  := ReadString(strSection, 'Caption', strSection);
      strCategory := ReadString(strSection, 'Category', '');
      // Custom Icons
      strIcon     := ReadString(strSection, 'Icon', DefaultIcon);
      if strIcon <> '' then
      begin
        strIcon := strDrive + StripDrive(strIcon);
      end;
      strOpenIcon := ReadString(strSection, 'OpenIcon', DefaultOpenIcon);
      if strOpenIcon <> '' then
      begin
        strOpenIcon := strDrive + StripDrive(strOpenIcon);
      end;
      intNew := AddCategory(strName, strCaption, strCategory, False, 0, strIcon, strOpenIcon);
      if (intNew <> -1) then
      begin
        Apps[intNew].SortOrder := ReadInteger('SortOrder', Apps[intNew].SortKey, 0);
      end;
      Inc(I);
      strSection := Format('Ctg_%d', [I]);
    end;
    Free;
  end;
end;

procedure TAppItemList.SaveCategories;
var
  I:          Integer;
  strSection: String;
  C:          Integer;
begin
  C := 0;
  with TIniFile.Create(DataFile) do
  begin
    for I := 0 to Count - 1 do
    begin
      if Apps[I].IsMenu then
      begin
        strSection := Format('Ctg_%d', [C]);
        EraseSection(strSection);
        WriteString(strSection, 'Name', Apps[I].Name);
        WriteString(strSection, 'Caption', Apps[I].Caption);
        if (Apps[I].CategoryName <> '') then
        begin
          WriteString(strSection, 'Category', Apps[I].CategoryName);
        end;
        if (Apps[I].ExePath <> '') and (Apps[I].ExePath <> DefaultIcon) then
        begin
          WriteString(strSection, 'Icon', StripDrive(Apps[I].ExePath));
        end;
        if (Apps[I].ExitExe <> '') and (Apps[I].ExePath <> DefaultOpenIcon) then
        begin
          WriteString(strSection, 'OpenIcon', StripDrive(Apps[I].ExitExe));
        end;
        Inc(C);
      end;
    end;
    while SectionExists(Format('Ctg_%d', [C])) do
    begin
      EraseSection(Format('Ctg_%d', [C]));
      Inc(C);
    end;
    Free;
  end;
end;

procedure TAppItemList.ChangeCtgIcon(NewIcon: String);
var
  I:        Integer;
  FileInfo: TSHFileInfo;
begin
  BeginUpdate;
  if NewIcon <> DefaultIcon then
  begin
    DefaultIcon := '';
    if FileExists(NewIcon) then
    begin
      DefaultIcon := NewIcon;
      for I := 0 to Count - 1 do
      begin
        if Apps[I].IsMenu then
        begin
          if (Apps[I].IconHandle <> 0) then
          begin
            DestroyIcon(Apps[I].IconHandle);
          end;
          FillChar(FileInfo, SizeOf(FileInfo), #0);
          SHGetFileInfo(PChar(NewIcon), 0, FileInfo, SizeOf(FileInfo),
            SHGFI_ICON or SHGFI_SMALLICON);
          Apps[I].IconHandle := FileInfo.hIcon;
        end;
      end;
    end;
  end;
  EndUpdate;
end;

procedure TAppItemList.ChangeOpenCtgIcon(NewIcon: String);
var
  I: Integer;
begin
  BeginUpdate;
  if (NewIcon <> DefaultOpenIcon) or (NewIcon = '') then
  begin
    DefaultOpenIcon := NewIcon;
    for I := 0 to Count - 1 do
    begin
      if Apps[I].IsMenu then
      begin
        Apps[I].AddOpenIcon(NewIcon);
      end;
    end;
  end;
  DefaultOpenIcon := NewIcon;
  EndUpdate;
end;

procedure TAppItemList.Apply(intIndex: Integer; sbtnTarget: TSpeedButton);
var
  brshIcon: hBrush;
  hndIcon:  THandle;
begin
  if (IsVisible(intIndex)) then
  begin
    hndIcon := Apps[intIndex].IconHandle;
    if Apps[intIndex].IsMenu then
    begin
      if Apps[intIndex].MenuOpen then
      begin
        sbtnTarget.Font := MenuOpenFont;
        if (Apps[intIndex].OpenIconHandle <> 0) then
        begin
          hndIcon := Apps[intIndex].OpenIconHandle;
        end;
      end else
      begin
        sbtnTarget.Font := MenuFont;
      end;
    end else
    begin
      sbtnTarget.Font := AppFont;
    end;
    sbtnTarget.Glyph.TransparentColor := TransColor;
    sbtnTarget.Margin  := Apps[intIndex].Level * MenuIndent;
    sbtnTarget.BringToFront;
    sbtnTarget.Caption := Apps[intIndex].Caption;
    sbtnTarget.Hint    := Apps[intIndex].Hint;
    brshIcon           := CreateSolidBrush(clWhite);
    sbtnTarget.Glyph.Canvas.Rectangle(0, 0, 16, 16);
    DrawIconEx(sbtnTarget.Glyph.Canvas.Handle, 0, 0, hndIcon,
      16, 16, 0, brshIcon, DI_NORMAL);
    sbtnTarget.Glyph.Canvas.Refresh;
    DeleteObject(brshIcon);
    sbtnTarget.Visible := True;
    sbtnTarget.BringToFront;
    sbtnTarget.Tag     := intIndex;
  end else
  begin
    sbtnTarget.Visible := False;
    sbtnTarget.Tag     := intIndex;
  end;
end;


procedure TAppItemList.DrawButtons(var Buttons: array of TSpeedButton);
var
  I:       Integer;
  intItem: Integer;
begin
  if Length(Buttons) > 0 then
  begin
    intItem := Buttons[0].Tag;
    if not IsVisible(intItem) then
    begin
      intItem := NextVisible(intItem);
    end;
    for I := 0 to Length(Buttons) - 1 do
    begin
      Apply(intItem, Buttons[I]);
      if IsVisible(intItem) then
      begin
        intItem := NextVisible(intItem);
      end;
    end;
  end;
end;

procedure TAppItemList.ShowSubItems(ItemIndex: Integer; var Buttons: array of TSpeedButton);
var
  I: Integer;
  P: Integer;
  C: Integer;
begin
  if not Apps[ItemIndex].IsMenu then
  begin
    Exit;
  end;
  if Buttons[Length(Buttons) - 1].Visible then
  begin
    // Last item is visible so we may have to scroll down
    if Buttons[0].Tag = ItemIndex then
    begin
      Exit;
    end;
    if not Apps[ItemIndex].IsMenu then
    begin
      Exit;
    end;
    if not Apps[ItemIndex].MenuOpen then
    begin
      Exit;
    end;
    C := 0;
    I := NextVisible(ItemIndex);
    while (I > 0) and (Apps[I].Level > Apps[ItemIndex].Level) do
    begin
      Inc(C);
      I := NextVisible(I);
    end;
    if (Length(Buttons) <= C + 1) then
    begin // Move Menu to top
      BeginUpdate;
      Buttons[0].Tag := ItemIndex;
      EndUpdate;
    end;
    P := -1;
    for I := 0 to Length(Buttons) - 1 do
    begin
      if Buttons[I].Tag = ItemIndex then
      begin
        P := I;
        Break;
      end;
    end;
    if P = -1 then
    begin
      Exit;
    end; // Menu top not visible
    if (Length(Buttons) - P > C) then
    begin
      Exit;
    end; // All Visible
    BeginUpdate;
    I := P + 1 - Length(Buttons) + C;
    if I > P then
    begin
      I := P;
    end;
    Scroll(Buttons, I);
    EndUpdate;
  end else
  begin
    // Last item isn't visible so we may have to scroll up
    if Apps[ItemIndex].MenuOpen then
    begin
      Exit;
    end;
    if FirstVisible = Buttons[0].Tag then
    begin
      Exit;
    end;
    if not Buttons[0].Visible then
    begin
      Exit;
    end;
    C := 0;
    I := Length(Buttons) - 1;
    while not Buttons[I + C].Visible do
    begin
      Dec(C);
    end;
    BeginUpdate;
    Scroll(Buttons, C);
    EndUpdate;
  end;
end;

procedure TAppItemList.FixLevels;
var
  I:      Integer;
  intCtg: Integer;
begin
  if (Count > 0) then
  begin
    BeginUpdate;
    fAppList.Sorted := False;
    for I := 0 to Count - 1 do
    begin
      fAppList[I] := Apps[I].GetSortString;
      if Apps[I].CategoryName = '' then
      begin
        Apps[I].Level := 0
      end else begin
        intCtg      := GetCtgByName(Apps[I].CategoryName);
        if intCtg >= 0 then
        begin
          Apps[I].Level := Apps[intCtg].Level + 1;
        end else
        begin
          Apps[I].Level := 0;
        end;
      end;
    end;
    fAppList.Sort;
    fAppList.Sorted := True;
    EndUpdate;
  end;
end;

function TAppItemList.PrepareSort(ItemIndex: Integer; slWork: TStringList): Integer;
  // Put All Items at this level in the menu into a stringlist in preparation for sort changes
var
  I: Integer;
  A: Integer;
begin
  Result := -1;
  slWork.Clear;
  I      := ItemIndex;
  while (I > 0) and (Apps[I - 1].Level >= Apps[ItemIndex].Level) do
  begin
    Dec(I);
  end; // Get First Item in Current Level
  while (I < Count) do
  begin
    if (Apps[I].Level = Apps[ItemIndex].Level) then
    begin
      A := slWork.AddObject(Apps[I].Caption, Apps[I]);
      if (I = ItemIndex) then
      begin
        Result := A;
      end;
    end else if (Apps[I].Level < Apps[ItemIndex].Level) then
    begin
      I := Count;
    end;
    Inc(I);
  end;
end;

function TAppItemList.CanSortUp(ItemIndex: Integer): Boolean;
var
  I: Integer;
begin
  Result := False;
  if (ItemIndex > 0) and (ItemIndex < Count) then
  begin
    I := ItemIndex - 1;
    while (I >= 0) and (not Result) and (Apps[I].Level >= Apps[ItemIndex].Level) do
    begin
      if Apps[I].Level = Apps[ItemIndex].Level then
      begin
        Result := True;
      end
      else
      begin
        Dec(I);
      end;
    end;
  end;
end;

function TAppItemList.CanSortDown(ItemIndex: Integer): Boolean;
var
  I: Integer;
begin
  Result := False;
  if (ItemIndex >= 0) and (ItemIndex < Count - 1) then
  begin
    I := ItemIndex + 1;
    while (I < Count) and (not Result) and (Apps[I].Level >= Apps[ItemIndex].Level) do
    begin
      if Apps[I].Level = Apps[ItemIndex].Level then
      begin
        Result := True;
      end
      else
      begin
        Inc(I);
      end;
    end;
  end;
end;

procedure TAppItemList.SortUp(ItemIndex: Integer);
var
  I:   Integer;
  L:   TStringList;
  App: TAppItem;
begin
  if (ItemIndex > 0) and (ItemIndex < Count) then
  begin
    L := TStringList.Create;
    I := PrepareSort(ItemIndex, L);
    if (I > 0) and (I < L.Count) then
    begin
      BeginUpdate;
      App          := TAppItem(L.Objects[I]);
      L.Objects[I] := L.Objects[I - 1];
      L.Objects[I - 1] := App;
      FinalizeSort(L);
      EndUpdate;
    end;
    L.Free;
  end;
end;

procedure TAppItemList.SortDown(ItemIndex: Integer);
var
  I:   Integer;
  L:   TStringList;
  App: TAppItem;
begin
  if (ItemIndex >= 0) and (ItemIndex < Count - 1) then
  begin
    L := TStringList.Create;
    I := PrepareSort(ItemIndex, L);
    if (I >= 0) and (I < L.Count - 1) then
    begin
      BeginUpdate;
      App          := TAppItem(L.Objects[I]);
      L.Objects[I] := L.Objects[I + 1];
      L.Objects[I + 1] := App;
      FinalizeSort(L);
      EndUpdate;
    end;
    L.Free;
  end;
end;

procedure TAppItemList.SortTop(ItemIndex: Integer);
var
  I:   Integer;
  L:   TStringList;
  App: TAppItem;
begin
  if (ItemIndex > 0) and (ItemIndex < Count) then
  begin
    L := TStringList.Create;
    I := PrepareSort(ItemIndex, L);
    if (I > 0) and (I < L.Count) then
    begin
      BeginUpdate;
      App := TAppItem(L.Objects[I]);
      while I > 0 do
      begin
        L.Objects[I] := L.Objects[I - 1];
        Dec(I);
      end;
      L.Objects[I] := App;
      FinalizeSort(L);
      EndUpdate;
    end;
    L.Free;
  end;
end;

procedure TAppItemList.SortBottom(ItemIndex: Integer);
var
  I:   Integer;
  L:   TStringList;
  App: TAppItem;
begin
  if (ItemIndex >= 0) and (ItemIndex < Count - 1) then
  begin
    L := TStringList.Create;
    I := PrepareSort(ItemIndex, L);
    if (I >= 0) and (I < L.Count - 1) then
    begin
      BeginUpdate;
      App := TAppItem(L.Objects[I]);
      while I < L.Count - 1 do
      begin
        L.Objects[I] := L.Objects[I + 1];
        Inc(I);
      end;
      L.Objects[I] := App;
      FinalizeSort(L);
      EndUpdate;
    end;
    L.Free;
  end;
end;

procedure TAppItemList.FinalizeSort(slWork: TStringList);
// Set the sortorder for items in sorting list, save to ini
var
  I: Integer;
  O: Integer;
begin
  with TIniFile.Create(DataFile) do
  begin
    O := 0;
    if TAppItem(slWork.Objects[0]).SortOrder <> 0 then
    begin
      DeleteKey('SortOrder', TAppItem(slWork.Objects[0]).SortKey);
    end;
    TAppItem(slWork.Objects[0]).SortOrder := 0;
    for I := 1 to slWork.Count - 1 do
    begin
      if CompareStr(TAppItem(slWork.Objects[I - 1]).Caption,
        TAppItem(slWork.Objects[I]).Caption) >= 0 then
      begin
        Inc(O);
      end;
      TAppItem(slWork.Objects[I]).SortOrder := O;
      if O = 0 then
      begin
        DeleteKey('SortOrder', TAppItem(slWork.Objects[I]).SortKey);
      end else
      begin
        WriteInteger('SortOrder', TAppItem(slWork.Objects[I]).SortKey, O);
      end;
    end;
    Free;
  end;
  FixLevels;  // reset sort strings
end;

procedure TAppItemList.Scroll(var Buttons: array of TSpeedButton; intSteps: Integer);
var
  I:          Integer;
  intItem:    Integer;
  intTopItem: Integer;
begin
  if Length(Buttons) > 0 then
  begin
    intTopItem := Buttons[0].Tag;
    intItem    := Buttons[0].Tag;
    I          := -1;
    while intSteps <> 0 do
    begin
      if intSteps < 0 then
      begin
        if not AtTop(intItem) then
        begin
          I := PrevVisible(intItem);
          Inc(intSteps);
        end else
        begin
          intSteps := 0;
        end;
      end else
      begin
        if not AtBottom(intItem, Length(Buttons)) then
        begin
          I := NextVisible(intItem);
          Dec(intSteps);
        end else
        begin
          intSteps := 0;
        end;
      end;
      if (I >= 0) then
      begin
        intItem := I;
      end;
    end;
    if intItem <> intTopItem then
    begin
      BeginUpdate;
      Buttons[0].Tag := intItem;
      EndUpdate; // put in to automate redrawing
    end;
  end;
end;

procedure TAppItemList.CategoryMenu(ItemIndex: Integer; mnuParent: TMenuItem);
var
  M: TMenuItem;
  I: Integer;
  P: Integer;
begin
  while (mnuParent.Count > 0) and (mnuParent.Items[0].Caption <> '-') do
  begin
    mnuParent.Delete(0);
  end;
  P := 0;
  mnuParent.Tag := ItemIndex;
  for I := 0 to Count - 1 do
  begin
    if (Apps[I].IsMenu) then
    begin
      M         := TMenuItem.Create(mnuParent);
      M.Caption := StringOfChar(' ', Apps[I].Level * 2) + Apps[I].Caption;
      M.Tag     := I;
      M.OnClick := mnuCategoryClick;
      M.Checked := Apps[I].Name = Apps[ItemIndex].CategoryName;
      mnuParent.Insert(P, M);
      Inc(P);
    end;
  end;
end;

procedure TAppItemList.mnuCategoryClick(Sender: TObject);
// Procedure linked to category menu items
var
  ItemIndex: Integer;
  CtgIndex:  Integer;
  I:         Integer;
begin
  CtgIndex := TControl(Sender).Tag;
  if CtgIndex >= 0 then
  begin
    ItemIndex := TMenuItem(Sender).Parent.Tag;
    if (ItemIndex >= 0) and (ItemIndex < Count) and (ItemIndex <> CtgIndex) then
    begin
      if TMenuItem(Sender).Checked then
      begin
        Apps[ItemIndex].CategoryName := '';
      end else
      begin
        I := CtgIndex;
        while (I <> -1) and (I <> ItemIndex) do
        begin
          I := GetCtgByName(Apps[I].CategoryName);
        end;
        if (I <> ItemIndex) then  // to prevent linking under own tree
        begin
          Apps[ItemIndex].CategoryName := Apps[CtgIndex].Name;
        end;
      end;
    end;
  end;
end;

procedure TAppItemList.BeginUpdate;
begin
  UpdateCount := fUpdateCount + 1;
end;

procedure TAppItemList.EndUpdate;
begin
  UpdateCount := fUpdateCount - 1;
end;

procedure TAppItemList.AutoExec;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
  begin
    if not Apps[I].IsMenu then
    begin
      if Apps[I].AutoExec then
      begin
        Apps[I].AppExecute;
      end;
    end;
  end;
end;

function TAppItemList.AppExecute(ItemIndex: Integer): Boolean;
begin
  Result := False;
  if (ItemIndex >= 0) and (ItemIndex < Count) then
  begin
    Result := Apps[ItemIndex].AppExecute;
  end;
end;

procedure TAppItemList.SetFontSize(intSize: Integer);
begin
  AppFont.Size      := intSize;
  MenuFont.Size     := intSize;
  MenuOpenFont.Size := intSize;
end;

procedure TAppItemList.ExitExec;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
  begin
    if not Apps[I].IsMenu then
    begin
      if Apps[I].ExecOnClose then
      begin
        Apps[I].ExitExecute;
      end;
    end;
  end;
end;

end.
